/* hstack.c  --  stack-based operations for HeightField - Lab  */
/*               JPB Jan. 13 1996                              */

#include <stdio.h>                      /* fopen() sprintf() etc. */
#include <stdlib.h>                     /* strtod() */
#include <string.h>                     /* strcpy() */
#include <ctype.h>                      /* tolower() */
#include <dirent.h>                     /* opendir, readdir */
#include "hl.h"

extern hfield *hst;          /* stack of heightfields */
extern int hp;                 /* pointer to top of stack: hst[hp] */
extern int stacksize;          /* total size of stack */

#define INCSTACK  if (++hp == stacksize) hp = 0     /* HF stack ptr++ */
#define DECSTACK  if (--hp < 0) hp = (stacksize-1)  /* HF stack ptr-- */
#define NHF       (PTYPE *)NULL                     /* null heightfield */

/* --------- functions defined here ---------- */
int hcopy(hfield *h1p,hfield h2);
int list_stack();

/* ----------------------- list contents of HF stack ------------------*/
int list_stack()
{
 int i,j,k,addsp;

 j = hp;
 for (i=0;i<stacksize;i++) {    /* scan through all HFs in stack */
   if (i==0) printf("X");        /* top three stack items are named */
   else if (i==1) printf("Y");
   else if (i==2) printf("Z");
   else printf("%d",i+1);
   if (hst[j].a == (PTYPE *)NULL) printf(": --- \n");
   else {
     printf(": %s ",hst[j].name);
     addsp = 15 - strlen(hst[j].name);
     for (k=0;k<addsp;k++) printf(" ");   /* add spaces to justify column */
     printf("[%d x %d]",hst[j].xsize,hst[j].ysize);
     printf(" range %1.2e .. %1.2e",hst[j].min,hst[j].max);
     if (hst[j].c) printf("  complex");
     printf("\n");
   }
   if (--j < 0) j = stacksize-1;  /* wrap around at end of stack array */
  }
  return(0);
}



int hpop()
{
 if (s_empty()) return(1);     /* check if stack is empty */
 free(hst[hp].a);               /* now guaranteed .a is not null */
 hst[hp].a = (PTYPE *)NULL;
 DECSTACK;
 return(0);
}

int hrot()
{int i, n, ohp;
 hfield tmp, tmp2;

 if (s_empty()) return(1);     /* check if stack is empty */

 ohp = hp;                     /* save old stack ptr */
 n = 0;
 for (i=0;(i<stacksize) && (hst[hp].a != NHF);i++) {
  DECSTACK;     /* move one backwards in stack */
  n++;
 }          /* n now equals number of non-null HFs on stack */
 INCSTACK;  /* sp now points at last non-null HF */
 tmp = hst[hp];               /* store HF at bottom of stack */

 hp = ohp;                     /* start back at top of stack */
 for (i=0;i<n;i++) {
   tmp2 = hst[hp];               /* pick up current HF in stack */
   hst[hp] = tmp;                /* replace with held HF */
   tmp = tmp2;                   /* swap helds */
   DECSTACK;                     /* move down stack one */
 }
 hp = ohp;                       /* return stack pointer to top */
 return(0);
}           /* hrot */

int hdup()
{
  hfield h1, h2;

 if (s_empty()) return(1);     /* check if stack is empty */
 h2 = hst[hp];
 h1.a = NULL;
 if (hcopy(&h1,h2) != 0) return(1);
 INCSTACK;
 if (hst[hp].a != NULL) free(hst[hp].a);
 hst[hp] = h1; 
 return(0);
} /* hdup() */

int hswap()
{ hfield tmp;   /* temporary storage for swap operation */
 int hp2;

 if (s_empty()) return(1);     /* check if stack is empty */
 hp2 = hp-1;
 if (hp2 < 0) hp2 = (stacksize-1);  /* previous element in stack */
 if ( hst[hp2].a == (PTYPE *)NULL ) {
   printf("Two elements required for swap operation.\n");
   return(1);
 }
 tmp = hst[hp];
 hst[hp] = hst[hp2];
 hst[hp2] = tmp;
 
 return(0);
} /* hswap() */

int hrot3()
{ hfield tmp2, tmp3;   /* temporary storage for swap operation */
 int hp2, hp3;

 if (stacksize < 3) {
  printf("Stack is too small for rot3.\n");
  return(1);
 }
 if (s_empty()) return(1);     /* check if stack is empty */
 hp2 = hp-1;
 if (hp2 < 0) hp2 = (stacksize-1);  /* previous element in stack */
 hp3 = hp2-1;
 if (hp3 < 0) hp3 = (stacksize-1);  /* previous element in stack */
 if (( hst[hp2].a == (PTYPE *)NULL ) || (hst[hp3].a==(PTYPE*)NULL)) {
   printf("Three elements required for rot3 operation.\n");
   return(1);
 }
 tmp2 = hst[hp2];
 tmp3 = hst[hp3];
 hst[hp2] = hst[hp];
 hst[hp3] = tmp2;
 hst[hp] = tmp3;
 
 return(0);
} /* hrot3() */

int hname(char *name)
{
 if (s_empty()) return(1);     /* check if stack is empty */
 strncpy(hst[hp].name,name,14);
 hst[hp].name[14] = '\0';             /* insure null termination */
 return(0);
} /* hname() */

           /* copy one HF structure in stack into another: h1 <- h2 */
int hcopy(hfield *h1p,hfield h2)
{
  int x,y;
  hfield h1;
  int xsize, ysize;

   h1 = *h1p;
   if ((h2.a) == (PTYPE *)NULL) return(1);  /* copying null hf? */
   xsize = h2.xsize;
   ysize = h2.ysize;
   if (h1.a != (PTYPE *)NULL) free(h1.a);
   if (h2.c) {if (h_newc(&h1,xsize,ysize,h2.name)) return(1); } 
   else { if (h_newr(&h1,xsize,ysize,h2.name)) return(1); }
   h1.min = h2.min;
   h1.max = h2.max;
   
   /*   printf("copying %s <- %s\n",h1.name,h2.name);  */
   if (h1.c) ysize *= 2;
   for (y=0;y<ysize;y++) {
     for (x=0;x<xsize;x++) {
      El(h1.a,x,y) = El(h2.a,x,y);
     }
   }
   *h1p = h1;
   return(0);
}  /* end hcopy() */

